@@ -171,14 +171,14 @@ static const char * const tag_name[] = {
 	"COPY",              /* "/"   47 0x2f 0b00101111 */
 	"WEAKEN",            /* "0"   48 0x30 0b00110000 */
 	"REGEXP",            /* "1"   49 0x31 0b00110001 */
-	"RESERVED_0",        /* "2"   50 0x32 0b00110010 */
-	"RESERVED_1",        /* "3"   51 0x33 0b00110011 */
-	"RESERVED_2",        /* "4"   52 0x34 0b00110100 */
-	"RESERVED_3",        /* "5"   53 0x35 0b00110101 */
-	"RESERVED_4",        /* "6"   54 0x36 0b00110110 */
-	"RESERVED_5",        /* "7"   55 0x37 0b00110111 */
-	"RESERVED_6",        /* "8"   56 0x38 0b00111000 */
-	"RESERVED_7",        /* "9"   57 0x39 0b00111001 */
+	"OBJECT_FREEZE",     /* "2"   50 0x32 0b00110010 */
+	"OBJECTV_FREEZE",    /* "3"   51 0x33 0b00110011 */
+	"RESERVED_0",        /* "4"   52 0x34 0b00110100 */
+	"RESERVED_1",        /* "5"   53 0x35 0b00110101 */
+	"RESERVED_2",        /* "6"   54 0x36 0b00110110 */
+	"RESERVED_3",        /* "7"   55 0x37 0b00110111 */
+	"RESERVED_4",        /* "8"   56 0x38 0b00111000 */
+	"RESERVED_5",        /* "9"   57 0x39 0b00111001 */
 	"FALSE",             /* ":"   58 0x3a 0b00111010 */
 	"TRUE",              /* ";"   59 0x3b 0b00111011 */
 	"MANY",              /* "<"   60 0x3c 0b00111100 */
